%%% Thomas A. Adams II
%%% thomas.a.adams@ntnu.no
%%% June 2020 Release
%%% This code is given without warranty or license, provided AS-IS
%%% Please site the corresponding academic journal article when citing and using this code:
%%%
%%% Christian DA, Adams TA II, et al. cDC1 coordinate innate and adaptive responses in the omentum required 
%%% for T cell priming and memory. Science Immunology 7, eabq7432 (2022). 
%%% https://doi.org/10.1126/sciimmunol.abq7432
%%%
%%% That paper is the documentation. This code is not actively maintained and may have incomplete parts.
%%%
%%% This contains the probability that a transition will occur given the age of the cell in hours. See paper for explanation.


% This function does a univariate search optimization on one specific variable
% Because my personal computer can have six workers in parallel, this is what it will
% do. This optimized for my specific computer. If you have less than three workers, then you will need to change the code. 
% If you have more workers, then it will do some additional random explorations using your available CPUunits to help explore faster.
% Worker 1: The point predicted by the newton method (numerical newton)
% Worker 2: A second point struck close to it
% Worker 3: A third point struck close to it
% Workers 4+: Points randomly to the sides of the current known global best

% maxrange and minrange should be 1xN row vectors. gbestx is [x1 x2 .. xN]
% row vector (just like PSO) this is for convenience
% VarNum is the variable number we are messing with
function [ gbestx, gbestf, conv, iters] = Univariate( fun, maxit, Np, maxrange, minrange, initial, VarNum, verbose )

% Get the lower and upper bounds of the variable we are messing with
% This is used for bisection
LB = minrange(VarNum);
UB = maxrange(VarNum);

if (Np < 3)
  fprintf('Error: Must have at least Np=3 points per loop.\n');
  return;
end

N = size(maxrange,2); % number of dimensions
if (size(minrange,2) ~= N)
  fprintf('Error: Dimensions of maxrange and minrange must be equal.\n');
  return;
end   

if (VarNum > N)
  fprintf('Error: Cannot vary something beyond dimensions of the variable (VarNum too high).\n');
  return;
end   


h = 0.05; % percentage of total range used for finite difference purturbations
z = 0.10; % percentage of total range used for exploring around the minimum
s = 0.10; % steepest descent starting stepsize factor
d = 0;    % direction of current travel +1 for upward, -1 for downard

% Set the initial points for each provided point
x = zeros(Np, N);
x(1,:) = initial; % point 1 is the initial point except
x(2,:) = initial; % point 1 is the initial point except perturbed
x(2,VarNum) = x(2,VarNum) - h * (UB - LB);
x(3,:) = initial; % point 1 is the initial point except perturbed. Don't worry if outside the bound
x(3,VarNum) = x(3,VarNum) + h * (UB - LB);

if (Np>=4) % 
  x(4,:) = initial; 
  x(4,VarNum) = UB;
end
if (Np>=5) % 
  x(5,:) = initial; 
  x(5,VarNum) = LB;
end

if (Np>=6) % if we have additional points, then just space them evenly in between the current bounds.
  Nextra = Np - 5;
  i = 0;
  for (p = 6:Np)
     i = i + 1;     
     x(p,:) = initial; 
     x(p,VarNum) = LB + i * (UB - LB) / (Nextra+1);      
  end
end

% Now run the loop. 
% the locations and values of the personal bests are just their current positions
gbestf = Inf;
NewtonPointsVisited = [];
BestPerturbPointsVisited = [];

% evaluate the function at each x. also keep track of which is the global best
%
for i=1:maxit
  % evaluate the functinos at each point
  prevx = x;
  prevbestf = gbestf;
        
  parfor k=1:Np
    ftemp(k) = fun(x(k,:));
    if (verbose)
      fprintf('Complete iteration %d, Particle %d, Variable %d, at location %1.5f, Function: %1.2f\n', i,k,VarNum,x(k,VarNum),ftemp(k));          
    end
  end
  % figure out which is the best
  for k=1:Np % so fast prolly better to leave it serial.       
    if (ftemp(k) < gbestf && x(k,VarNum) <= UB && x(k,VarNum) >= LB) % make sure its within bounds or else don't count it.
      gbestx = x(k,:);
      gbestf = ftemp(k);
    end
  end
  
  fprintf('Gbest is %1.4f at %1.4f\n', gbestf, gbestx(VarNum));
    
  for k=1:Np
    if (k <= 3)
      NewtonPointsVisited = [NewtonPointsVisited; [x(k, VarNum) ftemp(k)]];
    else
      BestPerturbPointsVisited = [BestPerturbPointsVisited; [x(k, VarNum) ftemp(k)]];
    end
  end    
    
  % First, steepest descent method. Use the first three points to create a second order centered finite differences derivative
  cfd = (ftemp(3) - ftemp(2)) / (2*h);
  dnew = -sign(cfd);
  if (dnew ~= d && d~=0) % if changed, drop the stepsize
    s = s * 0.5;
    h = h * 0.5;
  end
  d = dnew;
  
  % move the point and two around it for next iteration according to the
  % direction
  x(1,VarNum) = x(1,VarNum) + d * s * (UB - LB);
  % bounds check
  if (x(1,VarNum) < LB)
    x(1,VarNum) = LB;
  elseif (x(1,VarNum) > UB)
    x(1,VarNum) = UB;
  end
  x(2,VarNum) = x(1,VarNum) - h * (UB - LB);
  x(3,VarNum) = x(1,VarNum) + h * (UB - LB);
    
  % explore around our current global best randomly  
  if (Np >= 4)
    for k = 4:Np
      x(k,VarNum) = gbestx(VarNum) + 2 * z * (UB - LB) * (rand(1) - 0.5);
      if (x(k,VarNum) < LB)
        x(k,VarNum) = LB;
      elseif (x(k,VarNum) > UB)
        x(k,VarNum) = UB;
      end
    end
  end 
  
  

end
conv = 0;
iters = i;

%%% TEMP
% plot(NewtonPointsVisited(:,1), NewtonPointsVisited(:,2),'bo');
% hold on;
% plot(BestPerturbPointsVisited(:,1), BestPerturbPointsVisited(:,2),'ro');
% pause

end
