
%%% Thomas A. Adams II
%%% thomas.a.adams@ntnu.no
%%% June 2020 Release
%%% This code is given without warranty or license, provided AS-IS
%%% Please site the corresponding academic journal article when citing and using this code:
%%%
%%% Christian DA, Adams TA II, et al. cDC1 coordinate innate and adaptive responses in the omentum required 
%%% for T cell priming and memory. Science Immunology 7, eabq7432 (2022). 
%%% https://doi.org/10.1126/sciimmunol.abq7432
%%%
%%% That paper is the documentation. This code is not actively maintained and may have incomplete parts.
%%%
%%% This contains the code that runs a particle swarm optimization of a generic function. It has the option to use parallel computing to speed things up.


% maxit = maximum number of iterations
% Np = number of particles
% maxrange and minrange should be 1xN row vectors. gbestx is [x1 x2 .. xN] row vector
% assumes the function is a minimization function.
% gbestx is the row vector of the best found x
% gbestf is the objective function of that x
% conv is the normalized cloud size (i.e. shows how well the particles have
% converged). Smaller means very well converged.
% iters = actual number of iterations

function [ gbestx, gbestf, conv, iters] = pso( fun, maxit, Np, maxrange, minrange, initial, parallelize, verbose, VerboseFileID, stoponzero )


% Accepts a function to use as a black box objective function
% Accepts a range of the variables
% Optional: Takes one point (initial) and assigns it to one of the PSO variables
% all remaining variables are assigned through latin hypercube sampling
% initially
% optional: If you want to parallel computing it, say you do (better for
% long functino evals, but not for short ones)
% 
% purposes
if nargin <= 2
  fprintf('Error: First three function arguments are not optional.\n');
  return;
end

N = size(maxrange,2); % number of dimensions
if (size(minrange,2) ~= N)
  fprintf('Error: Dimensions of maxrange and minrange must be equal.\n');
  return;
end   
    
% Np = 30; % number of particles -- just go with 30 for now
w1 = 2.8; % classic parameters
w2 = 1.3; % classic parameters
chi = -2 / (2-w1-w2-sqrt((w1+w2)^2 - 4*(w1+w2)));
x = zeros(Np, N);

% check if we have an initial point
if nargin >= 6
  if (size(initial,2) ~= N)  
    fprintf('Error: Dimensions of maxrange and initial must be equal.\n');
    return;
  else
    x(1,:) = initial; 
    
    % if the initial point has anything outside the bounds, move it into
    % the bounds.
    
    for d = 1:N
      x(1,d) = min(x(1,d), maxrange(d));    
      x(1,d) = max(x(1,d), minrange(d));
    end
    
    NpToInitialize = Np - 1;
    startindex = 2;
  end
else
  NpToInitialize = Np;
  startindex = 1;
end   

% check if we have parallelization
if (nargin >= 7)
  if (parallelize ~= 0)
    parallelize = 1;
  else
    parallelize = 0;
  end
else
  parallelize = 0;
end

if (nargin >= 8)
  if (verbose ~= 0)
    verbose = 1;
  else
    verbose = 0;
  end
else
  verbose = 0;
end

if (nargin >= 9)
  if (VerboseFileID  ~= 0)
    printtofile  = 1;    
  else
    printtofile  = 0;
  end
else
  printtofile = 0;
end

if (nargin < 10)
  stoponzero = false;
end

% initialize all remaining particles
% use latin hypercube
for d=1:N % for each dimension
  % create the random LHC for that dimension
  pts = (maxrange(d)-minrange(d))*(randperm(NpToInitialize)-1)/(NpToInitialize-1) + minrange(d);
  % store in the x matrix in the appropriate column.
  x(startindex:Np,d) = pts';
end

% initialize velocities randomly (up to 25% of the range in each dimension)
v = (rand(Np, N)-0.5)*0.5;
for d=1:N % for each dimension
  v(:,d) = v(:,d) * (maxrange(d) - minrange(d));  
end

% the locations and values of the personal bests are just their current positions
pbestx = x;
pbestf = ones(Np,1) * Inf;
gbestf = Inf;

% evaluate the function at each x. also keep track of which is the global best
%
for i=1:maxit
    prevx = x;
    if (parallelize==1)
      parfor k=1:Np
        ftemp(k) = fun(x(k,:));
        if (verbose)
          fprintf('Complete iteration %d, Particle %d, Function: %1.2f\n', i,k,ftemp(k));          
        end
      end
    else
      for k=1:Np
        ftemp(k) = fun(x(k,:));
        if (verbose)
          fprintf('Complete iteration %d, Particle %d, Function: %1.2f\n', i,k,ftemp(k));
        end
        if (printtofile)
          fprintf(VerboseFileID, 'Complete iteration %d, Particle %d, Function: %1.2f\n', i,k,ftemp(k));
        end
      end  
    end
    for k=1:Np % so fast prolly better to leave it serial.       
      if (ftemp(k) < pbestf(k))
        pbestx(k,:) = x(k,:);
        pbestf(k) = ftemp(k);
        if (ftemp(k) < gbestf)
          gbestx = x(k,:);
          gbestf = ftemp(k);
        end
      end
    end

    % update velocities
    for d=1:N % for each dimension
      gbestvect(:,d) = gbestx(d) * ones(Np,1);      
    end 
    
    w1vec = rand(Np,1);
    w2vec = rand(Np,1);
    v = chi*(v + w1 * w1vec .* (pbestx - x) + w2 * w2vec .* (gbestvect - x));
    % do the move
       
    x = x + v;
    % enforce the bounds (use sticky boundaries, leave velocities intact)
    for d=1:N % for each dimension
      x(:,d) = min(x(:,d), maxrange(d));    
      x(:,d) = max(x(:,d), minrange(d));
    end
    
    % check for early exit
    DistancesToBestPoint = (x - gbestx);
    Normalizations = (maxrange-minrange);
    for d=1:N
      if (abs(Normalizations(d)) < 1e-10) % if the range is tiny, then just set the normalizer to 1. avoid weirdness
        Normalizations(d) = 1;
      end
    end
    NormalizedDistancesToBestPoint = DistancesToBestPoint ./ Normalizations;
    SquaredNormalizedDistancesToBestPoint = NormalizedDistancesToBestPoint .^ 2;
    SumOfSquaredNormalizedDistancesToBestPoint = sqrt(sum(SquaredNormalizedDistancesToBestPoint'));
    conv = norm(SumOfSquaredNormalizedDistancesToBestPoint);
    if (verbose)
      fprintf('Normalized Cloud Size: %1.4f\n', conv); 
    end
    
    if (conv<1e-3)
        fprintf('stopping early (%d iterations) since particles have concensus\n', i);        
        break;
    end
    
    % debug... early exit on known minimum (0)
    if (stoponzero == true && (abs(gbestf) < 0.01))
        fprintf('stopping early (%d iterations) since minimum reached or close\n', i);        
        break;
    end
    
end

% at the end, print out how close the points are to the best (i.e.
% concensus)

iters = i;
end

